"""
Train an autoencoder to reduce dimensionality of input features.
"""

import logging
import math
import os
import pickle
from collections.abc import Sequence
from typing import List

import numpy as np
import skfuzzy as fuzz
import tonik
import torch
import torch.nn as nn
import torch.nn.functional as F
import xarray as xr
import yaml
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import DataLoader, random_split

from zizou import AnomalyDetectionBaseClass, SliceBatchSampler, ZizouDataset

logger = logging.getLogger(__name__)


class TiedAutoEncoder(nn.Module):
    """
    Applies an tied-weight autoencoder to the incoming data.
    From https://gist.github.com/northanapon/375e17fb395391c144deff20914e51df
    Parameters
    -------------
    in_features : int
        size of each input sample.
    h_features : List[int]
        a list of size of each layer for the encoder and
        the reverse for the decoder.
    activation : function, optional
        an activation function to apply for each layer
        (the default is torch.tanh).
    """

    def __init__(self, in_features: int, h_features: List[int], activation=F.relu):
        """Create an autoencoder."""
        super().__init__()
        self.in_features = in_features
        self.h_features = h_features
        self.encoded_features = h_features[-1]
        self.activation = activation
        self.weights = nn.ParameterList([])
        self.enc_biases = nn.ParameterList([])
        self.dec_biases = nn.ParameterList([])
        in_dim = in_features
        for h in h_features:
            self.weights.append(nn.Parameter(torch.DoubleTensor(h, in_dim)))
            self.enc_biases.append(nn.Parameter(torch.DoubleTensor(h)))
            in_dim = h
        for h in reversed(h_features[:-1]):
            self.dec_biases.append(nn.Parameter(torch.DoubleTensor(h)))
        self.dec_biases.append(nn.Parameter(torch.DoubleTensor(in_features)))
        self.reset_parameters()
        self.loss = torch.nn.MSELoss(reduction="none")

    def forward(self, x: torch.DoubleTensor):
        """Return result of encoding and decoding."""
        dec = self.encode(x)
        return self.decode(dec)

    def encode(self, x: torch.DoubleTensor):
        """Return encoded data."""
        o = x
        for w, b in zip(self.weights[:-1], self.enc_biases[:-1]):
            o = F.linear(o, w, b)
            o = self.activation(o)
        return F.linear(o, self.weights[-1], self.enc_biases[-1])

    def decode(self, o: torch.DoubleTensor):
        """Return decoded data."""
        r_weights = list(reversed(self.weights))
        for w, b in zip(r_weights[:-1], self.dec_biases[:-1]):
            o = F.linear(o, w.t(), b)
            o = self.activation(o)
        return F.linear(o, r_weights[-1].t(), self.dec_biases[-1])

    def reset_parameters(self):
        """Reset linear module parameters (from nn.Linear)."""
        for w, b in zip(self.weights, self.enc_biases):
            nn.init.kaiming_uniform_(w, a=math.sqrt(5))
            if b is not None:
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w)
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(b, -bound, bound)
        for w, b in zip(self.weights, self.dec_biases):
            if b is not None:
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w.t())
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(b, -bound, bound)


class FuzzyCluster(TransformerMixin, BaseEstimator):
    def __init__(self, n_clusters=5, error=0.005, maxiter=1000, init=None):
        self.n_clusters_ = n_clusters
        self.error_ = error
        self.maxiter_ = maxiter
        self.init_ = init
        self.clusterscaler = MinMaxScaler()
        self._is_fitted = False

    def fit(self, data, scalerfile, clustercenterfile):
        try:
            with open(scalerfile, "rb") as fh:
                self.clusterscaler = pickle.load(fh)
            self.cntr_ = np.load(clustercenterfile)
            self._is_fitted = True
            logging.info("Loaded cluster model.")
        except FileNotFoundError:
            data_scaled = self.clusterscaler.fit_transform(data)
            self.cntr_, u, u0, d, jm, p, fpc = fuzz.cluster.cmeans(
                data_scaled.T, self.n_clusters_, 2, error=0.005, maxiter=1000, init=None
            )
            with open(scalerfile, "wb") as fh:
                pickle.dump(self.clusterscaler, fh)
            np.save(clustercenterfile, self.cntr_)
            self._is_fitted = True

    def transform(self, data):
        assert self._is_fitted
        data_scaled = self.clusterscaler.transform(data)
        u, u0, d, jm, p, fpc = fuzz.cluster.cmeans_predict(
            data_scaled.T, self.cntr_, 2, error=0.005, maxiter=1000
        )
        return u

    def fit_transform(self, data, scalerfile, clustercenterfile):
        self.fit(data, scalerfile, clustercenterfile)
        return self.transform(data)


class AutoEncoder(AnomalyDetectionBaseClass):
    features = [
        "rsam",
        "dsar",
        "central_freq",
        "predom_freq",
        "bandwidth",
        "rsam_energy_prop",
        "sonogram",
    ]

    transform_dict = {"rsam": "log"}

    stack_dict = {
        "dsar": "2D",
        "central_freq": "1h",
        "predom_freq": "1h",
        "variance": "1h",
        "bandwidth": "1h",
    }

    files = dict(
        modelfile="autoencoder_weights.pth",
        scalerfile="cluster_scaler.pkl",
        clustercenterfile="cluster_center.npy",
    )

    def __init__(
        self,
        store: tonik.Storage,
        features=None,
        layers: Sequence = [2000, 500, 200, 6],
        epochs: int = 5,
        batch_size: int = 1024,
        patience: int = 10,
        device: str = "cpu",
        n_clusters: int = 5,
        configfile: str = None,
    ):
        self.store = store
        if features is not None:
            self.features = features
        self.layers_ = layers
        self.epochs_ = epochs
        self.batch_size_ = batch_size
        self.patience_ = patience
        self.device_ = device
        if configfile is not None:
            try:
                with open(configfile, "r") as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.layers_ = c["autoencoder"].get("layers", layers)
            self.epochs_ = c["autoencoder"].get("epochs", epochs)
            self.patience_ = c["autoencoder"].get("patience", patience)
            self.device_ = c["autoencoder"].get("device", device)

        super(AutoEncoder, self).__init__(
            self.features, self.stack_dict, self.transform_dict
        )
        self.device_ = (
            self.device_
            if torch.cuda.is_available()
            else "mps"
            if torch.backends.mps.is_available()
            else "cpu"
        )
        logger.info(f"Using {self.device_} device")
        self.cluster_ = FuzzyCluster(n_clusters=n_clusters)
        self.classifications = None
        self.dataset = ZizouDataset(
            store, self.features, target_transform=torch.from_numpy
        )
        autoencoder_ = TiedAutoEncoder(self.dataset.nfeatures, self.layers_)
        self.model_ = autoencoder_.to(self.device_)
        logger.info(self.model_)

    def clear(self):
        for _f in self.files.values():
            try:
                os.remove(os.path.join(self.store.path, _f))
            except FileNotFoundError:
                pass

    def checkpoint(self, model, filename):
        torch.save(model.state_dict(), filename)

    def resume(self, model, filename):
        model.load_state_dict(torch.load(filename))

    def train(self, dataloader, optimizer):
        size = len(dataloader.dataset)
        self.model_.train()
        for batch, (X, y) in enumerate(dataloader):
            X, y = X.to(self.device_), y.to(self.device_)

            # Compute prediction error
            pred = self.model_(X)
            loss = self.model_.loss(pred, y).sum(axis=1).mean()

            # Backpropagation
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            if batch % 100 == 0:
                loss, current = loss.item(), (batch + 1) * len(X)
                logger.info(f"Training loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

    def test(self, dataloader, loss_fn):
        num_batches = len(dataloader)
        self.model_.eval()
        test_loss = 0
        with torch.no_grad():
            for X, y in dataloader:
                X, y = X.to(self.device_), y.to(self.device_)
                pred = self.model_(X)
                test_loss += loss_fn(pred, y).sum(axis=1).mean()
        test_loss /= num_batches
        self.loss_ = test_loss
        logger.info(f"Test loss: {test_loss:>8f} \n")
        return test_loss

    def _is_fitted(self, store):
        try:
            self.model_.load_state_dict(
                torch.load(os.path.join(store.path, self.files["modelfile"]))
            )
            logging.info("Loaded pretrained model.")
        except FileNotFoundError:
            return False
        return True

    def fit(self, new=False):
        if self._is_fitted(self.store) and not new:
            return
        optimizer = torch.optim.SGD(self.model_.parameters(), lr=1e-3)
        logger.info("Splitting dataset...")
        train_size = int(0.8 * len(self.dataset))
        train_sampler = SliceBatchSampler(range(train_size), self.batch_size_, False)
        test_sampler = SliceBatchSampler(
            range(train_size, len(self.dataset)), self.batch_size_, False
        )
        train_dataloader = DataLoader(self.dataset, batch_sampler=train_sampler)
        test_dataloader = DataLoader(self.dataset, batch_sampler=test_sampler)
        logger.info("Starting training...")
        best_loss = 1e30
        best_epoch = 0
        for i in range(self.epochs_):
            logger.info(f"Epoch {i+1}")
            self.train(train_dataloader, optimizer)
            loss = self.test(test_dataloader, self.model_.loss)
            if loss < best_loss:
                best_loss = loss
                best_epoch = i
                self.checkpoint(
                    self.model_, os.path.join(self.store.path, self.files["modelfile"])
                )
            elif i - best_epoch > self.patience_:
                logger.info(f"Early stopped training at epoch {i+1}")
                break

    def transform(self, starttime, endtime):
        assert self._is_fitted(self.store)

        logger.info("Starting transform process.")
        self.store.starttime = starttime
        self.store.endtime = endtime
        zd = ZizouDataset(self.store, self.features, target_transform=torch.from_numpy)
        sampler = SliceBatchSampler(range(len(zd)), self.batch_size_, False)
        dataloader = DataLoader(zd, batch_sampler=sampler)
        logger.info("Transforming data.")
        with torch.no_grad():
            for X, y in dataloader:
                X = X.to(self.device_)
                encoded_data = self.model_.encode(X)
                X_hat = self.model_.decode(encoded_data)

                encoded_data = encoded_data.cpu().numpy()

                output = {}
                dates = zd.dates
                logger.info("Calculating loss between input and decoded output.")
                loss = self.model_.loss(X, X_hat).sum(axis=1)
                output["autoencoder_loss"] = xr.DataArray(
                    loss.cpu().numpy(), coords=[dates], dims=["datetime"]
                )
                logger.info(
                    f"Clustering embeddings into {self.cluster_.n_clusters_} cluster."
                )
                clustered_data = self.cluster_.fit_transform(
                    encoded_data,
                    os.path.join(self.store.path, self.files["scalerfile"]),
                    os.path.join(self.store.path, self.files["clustercenterfile"]),
                )
                cluster_names = list(range(self.cluster_.n_clusters_))
                output["autoencoder_cluster"] = xr.DataArray(
                    clustered_data,
                    coords=[cluster_names, dates],
                    dims=["cluster", "datetime"],
                )
                output["autoencoder_embedding"] = xr.DataArray(
                    encoded_data.T,
                    coords=[list(range(encoded_data.shape[1])), dates],
                    dims=["autodim", "datetime"],
                )
                xds = xr.Dataset(output)
                xds.attrs["station"] = self.store.station
                self.store.save(xds)

    def fit_transform(self, starttime, endtime):
        self.fit()
        self.classifications = self.transform(starttime, endtime)
        return self.classifications



import importlib
from os import PathLike
from typing import Optional

from .zizou_logging import setup_logging

setup_logging()

from .anomaly_base import (
    AnomalyDetectionBaseClass,
    SliceBatchSampler,
    ZizouDataset,
    min_max_scale_netcdf,
)
from .autoencoder import AutoEncoder
from .data import DataSource, FDSNWaveforms, MockSDSWaveforms, S3Waveforms, SDSWaveforms
from .dsar import DSAR
from .feature_base import FeatureBaseClass
from .rsam import RSAM, EnergyExplainedByRSAM
from .spectral_features import SpectralFeatures
from .ssam import SSAM

__all__ = [
    "AnomalyDetectionBaseClass",
    "FeatureRequest",
    "FeatureBaseClass",
    "RSAM",
    "SpectralFeatures",
    "SSAM",
    "get_data",
]


def get_data(filename: Optional[PathLike] = None) -> str:
    """Return path to zizou package.

    Parameters
    ----------
    filename : Pathlike, default None
        Append `filename` to returned path.

    Returns
    -------
    pkgdir_path

    """
    f = importlib.resources.files(__package__)
    return str(f) if filename is None else str(f / filename)



import datetime
import json
import logging
import os

import h5netcdf
import numpy as np
import pandas as pd
import torch
import xarray as xr
from torch.utils.data import BatchSampler, Dataset

logger = logging.getLogger(__name__)


def min_max_scale_netcdf(
    input_file, variable_name, scaling_params_file, chunk_size=144
):
    if os.path.isfile(scaling_params_file):
        with open(scaling_params_file, "r") as f:
            scaling_params = json.load(f)
        return scaling_params

    with xr.open_dataset(input_file, group="original") as ds:
        data = ds[variable_name]
        if len(data.sizes) > 1:
            min_value = list(data.min(axis=1).values)
            max_value = list(data.max(axis=1).values)
            median_value = list(data.median(axis=1).values)
        else:
            min_value = float(data.min())
            max_value = float(data.max())
            median_value = float(data.median())
    scaling_params = {"min": min_value, "max": max_value, "median": median_value}
    with open(scaling_params_file, "w") as f:
        json.dump(scaling_params, f)

    return scaling_params


class SliceBatchSampler(BatchSampler):
    def __init__(self, sampler, batch_size, drop_last):
        super().__init__(sampler, batch_size, drop_last)
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        # Generate all indices from the sampler
        indices = list(self.sampler)

        # Yield slices instead of list of indices
        for i in range(indices[0], indices[-1] + 1, self.batch_size):
            if self.drop_last and i + self.batch_size > indices[-1] + 1:
                break
            yield slice(i, min(i + self.batch_size, indices[-1] + 1))


class ZizouDataset(Dataset):
    def __init__(self, store, features, target_transform=None, feature_transforms=None):
        self.target_transform = target_transform
        self.feature_transforms = feature_transforms
        self.store = store
        self.features = features
        self.resolution = None
        self.starttime = store.starttime
        self.endtime = store.endtime
        self.init_dataset()
        self.npts = int(
            (self.endtime - self.starttime) / datetime.timedelta(hours=self.resolution)
        )
        self.nfeatures = self.get_shapes()[1]
        self.dates = None

    def get_dates(self):
        return pd.date_range(
            start=self.store.starttime,
            end=self.store.endtime,
            freq=datetime.timedelta(hours=self.resolution),
        )

    def init_dataset(self):
        self.scaling_params = {}
        for feature in self.features:
            scaling_params_file = os.path.join(
                self.store.path, f"scaling_params_{feature}.json"
            )
            fin = self.store.feature_path(feature)
            self.scaling_params[feature] = min_max_scale_netcdf(
                fin, feature, scaling_params_file
            )
            self.get_resolution(fin)

    def get_resolution(self, fin):
        with h5netcdf.File(fin, "r") as f:
            _resolution = f["original"].attrs["resolution"]
            if self.resolution is None:
                self.resolution = _resolution
            else:
                if self.resolution != _resolution:
                    msg = "Resolution is not consistent.\n"
                    msg += f"Expected {self.resolution}, got {_resolution}.\n"
                    msg += f"Feature: {fin}\n"
                    raise ValueError(msg)

    def get_shapes(self):
        nsamples = None
        nfeatures = 0
        for feature in self.features:
            shape_dict = self.store.shape(feature)
            for key, value in shape_dict.items():
                if key != "datetime":
                    nfeatures += value
                else:
                    if len(shape_dict) == 1:
                        nfeatures += 1
                    if nsamples is None:
                        nsamples = value
                    else:
                        if nsamples != value:
                            msg = "Number of samples is not consistent."
                            msg += f"Expected {nsamples}, got {value}."
                            raise ValueError(msg)
        return (nsamples, nfeatures)

    def load_and_scale_feature(self, feature):
        data = self.store(feature)
        if len(data.shape) > 1:
            data = data.fillna(
                np.array(self.scaling_params[feature]["median"])[:, np.newaxis]
            )
            data = (
                data - np.array(self.scaling_params[feature]["min"])[:, np.newaxis]
            ) / (
                np.array(self.scaling_params[feature]["max"])[:, np.newaxis]
                - np.array(self.scaling_params[feature]["min"])[:, np.newaxis]
            )
        else:
            data = data.fillna(self.scaling_params[feature]["median"])
            data = (data - self.scaling_params[feature]["min"]) / (
                self.scaling_params[feature]["max"]
                - self.scaling_params[feature]["min"]
            )
        return data

    def load_data(self, starttime, endtime):
        features = []
        var_cnt = 0
        for _f in self.features:
            logger.debug(f"Reading feature {_f} between {starttime} and {endtime}.")
            self.store.starttime = starttime
            self.store.endtime = endtime
            feat = self.load_and_scale_feature(_f)
            if len(feat.shape) > 1:
                for c in range(feat.shape[0]):
                    da = feat[c].reset_coords(drop=True)
                    feat_name = f"variable_{var_cnt}"
                    var_cnt += 1
                    features.append(da.rename(feat_name))
            else:
                if self.feature_transforms is not None:
                    features.append(
                        self.feature_transforms[_f](feat.rename(f"variable_{var_cnt}"))
                    )
                else:
                    features.append(feat.rename(f"variable_{var_cnt}"))
                var_cnt += 1
        feats = xr.merge(features).to_array()
        self.dates = feats.datetime.values
        return feats.transpose("datetime", "variable").values

    def __len__(self):
        return self.npts

    def __getitems__(self, idx):
        return self.__getitem__(idx)

    def __getitem__(self, idx):
        if isinstance(idx, int) or isinstance(idx, np.int64):
            start = [int(idx)]
            stop = [int(idx)]
        elif isinstance(idx, slice):
            start = idx.start
            stop = idx.stop
            if start is None:
                start = 0
            else:
                start = idx.start
            if stop is None:
                stop = self.npts - 1
            elif stop >= self.npts:
                stop = self.npts - 1
            else:
                stop = idx.stop - 1
            assert stop >= start, "Stop index must be greater than start index."
            start = [start]
            stop = [stop]

        elif isinstance(idx, list) or isinstance(idx, tuple):
            start = list(idx)
            stop = list(idx)

        elif isinstance(idx, np.ndarray):
            start = idx.tolist()
            stop = idx.tolist()

        # Handle PyTorch tensor of indices
        elif isinstance(idx, torch.Tensor):
            if (
                idx.dtype == torch.long or idx.dtype == torch.int
            ):  # Ensure it's an index tensor
                start = idx.tolist()
                stop = idx.tolist()
        else:
            raise TypeError(f"Invalid index type: {type(idx)}")

        all_data = []
        for _st, _sp in zip(start, stop):
            date_start = self.starttime + datetime.timedelta(
                hours=_st * self.resolution
            )
            date_end = self.starttime + datetime.timedelta(hours=_sp * self.resolution)
            _data = self.load_data(date_start, date_end)
            all_data.append(_data)
        sample = np.array(all_data).squeeze()
        target = np.array(all_data).squeeze()
        if self.target_transform is not None:
            sample = self.target_transform(sample)
            target = self.target_transform(target)
        return sample, target


class AnomalyDetectionBaseClass(object):
    def __init__(self, features, stacks=None, transforms=None):
        self.features = features
        self.stacks = stacks if stacks is not None else dict()
        self.transforms = transforms if transforms is not None else dict()

    def fit(self, starttime, endtime):
        """
        Train the model.
        """
        msg = "'training' is not implemented yet"
        raise NotImplementedError(msg)

    def infer(self, starttime, endtime):
        """
        Use the model to infer.
        """
        msg = "'inference' is not implemented yet"
        raise NotImplementedError(msg)

    def write_model_parameters(self):
        """
        Write principal components or
        model weights to file
        """
        msg = "'write_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)

    def read_model_parameters(self):
        """
        Read principal components or
        model weights from file
        """
        msg = "'read_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)

    def write_hyperparameters(self):
        """
        Store things like random seed,
        learning rate, number of principal
        components etc.
        """
        msg = "'write_hyperparameters' is not implemented yet"
        raise NotImplementedError(msg)

    def compute_anomaly_index(self):
        """
        A scalar value between 0 and 1 indicating
        the degree of anomaly (1=most anomaluous).
        """
        msg = "'compute_anomaly_index' is not implemented yet"
        raise NotImplementedError(msg)

    def get_features(self, data, featurefile=None):
        try:
            feats = xr.open_dataarray(featurefile)
        except (FileNotFoundError, ValueError):
            features = []
            for _f in self.features:
                logger.info(f"Reading feature {_f}.")
                feat = data(_f, self.stacks.get(_f, None))
                if _f == "sonogram":
                    for c in range(feat.shape[0]):
                        da = feat[c].reset_coords(names=["sonofrequency"], drop=True)
                        feat_name = "sonogram_{:d}".format(c)
                        features.append(da.rename(feat_name))
                else:
                    if self.transforms.get(_f, None) == "log":
                        features.append(np.log10(feat))
                    else:
                        features.append(feat)

            feats = xr.merge(features).to_array()
            if featurefile is not None:
                feats.to_netcdf(featurefile)
        return feats.transpose("datetime", "variable")



"""
Compute the Real-time Seismic-Amplitude Measurement (RSAM).
"""

import logging

import numpy as np
import pandas as pd
import scipy.integrate
import xarray as xr
import yaml
from obspy import UTCDateTime

from zizou import FeatureBaseClass
from zizou.util import (
    apply_freq_filter,
    round_time,
    trace_window_data,
    trace_window_times,
)

logger = logging.getLogger(__name__)

__all__ = ["RSAM", "EnergyExplainedByRSAM"]


def rsam(arr, axis=None):
    return np.absolute(arr).mean(axis=axis)


def rsam_wideband_energy_ratio(arr_rsam, arr_wideband, axis=-1):
    energy_rsam_band = scipy.integrate.simpson(arr_rsam**2, axis=axis)
    energy_wide_band = scipy.integrate.simpson(arr_wideband**2, axis=axis)
    return energy_rsam_band / energy_wide_band


class RSAM(FeatureBaseClass):
    """
    RSAM is the mean of the absolute value of a signal filtered
    between 2 and 5 Hz.

    :param interval: Length of the time interval in seconds over
                     which to compute RSAM.
    :type interval: int, optional
    :param filtertype: The type of filtering that is applied before
                       computing RSAM. Can be either band-pass ('bp'),
                       high-pass ('hp'), or low-pass ('lp').
    :type filtertype: str, optional
    :param filterfreq: The low and high cutoff frequency. If filtertype is
                       highpass or lowpass, only the first or last value is
                       used, respectively.

    :type filterfreq: tuple, optional
    """

    features = ["rsam"]

    def __init__(
        self,
        interval=600.0,
        filtertype="bandpass",
        filterfreq=(2, 5),
        configfile=None,
        reindex=True,
    ):
        super(RSAM, self).__init__()
        self.interval = interval
        self.filtertype = filtertype
        self.filterfreq = filterfreq
        self.feature = None
        self.reindex = reindex

        if configfile is not None:
            try:
                with open(configfile, "r") as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.interval = c["default"].get("interval", interval)
            cr = c.get("rsam")
            if cr is not None:
                self.filtertype = cr.get("filtertype", filtertype)
                freq = cr.get("filterfreq")
                if freq is not None:
                    self.filterfreq = (freq["low"], freq["high"])
                self.reindex = cr.get("reindex", reindex)

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        if len(trace) < 1:
            raise ValueError("Trace is empty.")

        logger.info(
            "Computing RSAM for {} between {} and {}.".format(
                ".".join(
                    (
                        trace.stats.network,
                        trace.stats.station,
                        trace.stats.location,
                        trace.stats.channel,
                    )
                ),
                trace.stats.starttime.isoformat(),
                trace.stats.endtime.isoformat(),
            )
        )

        tr = trace.copy()
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime

        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr.data)
        tr.data = mdata
        st = tr.split()

        apply_freq_filter(st, self.filtertype, self.filterfreq)
        st.merge(fill_value=np.nan)
        st.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr = st[0]

        # initialise dataframe
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for trint in trace_window_data(tr, self.interval, min_len=0.8):
            absolute = np.absolute(trint.data)  # absolute value
            trint.data = absolute  # assign back to trace
            mean = trint.data.mean()
            # convert to nanometres so dealing with whole numbers
            feature_data.append(mean / 1e-9)
            feature_idx.append(trint.stats.starttime.strftime("%Y-%m-%dT%H:%M:%S"))

        xda = xr.DataArray(
            feature_data, coords=[pd.DatetimeIndex(feature_idx)], dims="datetime"
        )
        xdf = xr.Dataset({"rsam": xda})
        xdf.attrs["starttime"] = trace.stats.starttime.isoformat()
        xdf.attrs["endtime"] = trace.stats.endtime.isoformat()
        xdf.attrs["station"] = trace.stats.station
        if self.reindex:
            starttime = UTCDateTime(str(xda.datetime.data[0]))
            starttime = round_time(starttime, self.interval)
            endtime = UTCDateTime(str(xda.datetime.data[-1]))
            endtime = round_time(endtime, self.interval)
            new_index = pd.date_range(
                starttime.datetime, endtime.datetime, freq="%dS" % int(self.interval)
            )
            xdf = xdf.reindex(dict(datetime=new_index), method="nearest")
        self.feature = xdf
        return self.feature


class EnergyExplainedByRSAM(RSAM):
    """
    The proportion of signal energy in the RSAM
    bandwidth, relative to a 0.5-10 Hz bandwidth.
    """

    features = ["rsam_energy_prop"]

    def __init__(
        self,
        interval=600.0,
        filtertype=None,
        filterfreq=(None, None),
        filtertype_wb="bandpass",
        filterfreq_wb=(0.5, 10.0),
        configfile=None,
    ):
        super(EnergyExplainedByRSAM, self).__init__(
            interval=interval,
            filtertype=filtertype,
            filterfreq=filterfreq,
            configfile=configfile,
        )
        self.filtertype_wb = filtertype_wb
        self.filterfreq_wb = filterfreq_wb

        # base params already read from configfile during superclass init
        # reopen and extract wideband filter params if any
        if configfile is not None:
            try:
                with open(configfile, "r") as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            cr = c.get("rsam")
            if cr is not None:
                self.filtertype_wb = cr.get("filtertype_wb", filtertype_wb)
                freq = cr.get("filterfreq_wb")
                if freq is not None:
                    self.filterfreq_wb = (freq["low"], freq["high"])

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        if len(trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)

        logger.info(
            "Computing Energy explained by RSAM for {} between {} and {}.".format(
                ".".join(
                    (
                        trace.stats.network,
                        trace.stats.station,
                        trace.stats.location,
                        trace.stats.channel,
                    )
                ),
                trace.stats.starttime.isoformat(),
                trace.stats.endtime.isoformat(),
            )
        )

        tr = trace.copy()
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime

        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr.data)
        tr.data = mdata

        st_rsam = tr.split()
        st_wb = st_rsam.copy()

        apply_freq_filter(st_rsam, self.filtertype, self.filterfreq)
        st_rsam.merge(fill_value=np.nan)
        st_rsam.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_rsam = st_rsam[0]

        # Get trace in wide bandwidth
        apply_freq_filter(st_wb, self.filtertype_wb, self.filterfreq_wb)
        st_wb.merge(fill_value=np.nan)
        st_wb.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_wb = st_wb[0]

        # initialise dataframe
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for t0, t1 in trace_window_times(tr_rsam, self.interval, min_len=0.8):
            energy_ratio = rsam_wideband_energy_ratio(
                tr_rsam.slice(t0, t1, nearest_sample=False).data,
                tr_wb.slice(t0, t1, nearest_sample=False).data,
            )
            feature_data.append(energy_ratio)
            feature_idx.append(t0.strftime("%Y-%m-%dT%H:%M:%S"))

        xdf = pd.DataFrame(
            data=feature_data,
            index=pd.DatetimeIndex(feature_idx, name="datetime"),
            columns=["rsam_energy_prop"],
            dtype=float,
        ).to_xarray()

        xdf.attrs["starttime"] = trace.stats.starttime.isoformat()
        xdf.attrs["endtime"] = trace.stats.endtime.isoformat()
        xdf.attrs["station"] = trace.stats.station

        self.feature = xdf
        self.trace = tr
        return self.feature
